/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package characters.robot;

import app.Measure;
import characters.AbstractCharacter;
import characters.Player;
import com.jme3.material.Material;
import com.jme3.math.FastMath;
import com.jme3.math.Vector2f;
import com.jme3.math.Vector3f;
import com.jme3.scene.Geometry;
import com.jme3.scene.Mesh;
import com.jme3.scene.Node;
import com.jme3.scene.VertexBuffer;
import java.util.List;

/**
 *
 * @author peter
 */
class RelativePosition {

    private Vector2f v2fromTo;
    private final Vector2f north = new Vector2f(0, -1); //x=0,z=-1 is usually considered north in Unity (by default, the Terrain Toolkit / Wind widget shows North as negative Z, http://answers.unity3d.com/questions/41837/what-direction-is-north-in-the-editor.html)
    private float compassAngle;

    /// <summary>
    /// Creates new relative position from vectors representing point from and to.
    /// </summary>
    /// <param name='newV3from'>
    /// vector representing the point from 
    /// </param>
    /// vector representing the point from 
    /// New v3to.
    /// </param>
    public RelativePosition(Vector3f newV3from, Vector3f newV3to) {
        resetFromTo(newV3from, newV3to);
    }

    /// <summary>
    /// Creates new relative position from compass angle in degrees and minutes and distance.
    /// </summary>
    /// <param name='degrees'>
    /// Degrees of the compass angle. From 0 to 359
    /// </param>
    /// <param name='minutes'>
    /// Minutes of the compass angle. From 0 to 59
    /// </param>
    /// <param name='distanceCm'>
    /// Distance in cm.
    /// </param>
    public RelativePosition(int degrees, int minutes, int distanceCm) {
        resetDgrMinDstInCm(degrees, minutes, distanceCm);
    }

    /// <summary>
    /// Creates new relative position from compass angle in minutes and distance.
    /// </summary>
    /// <param name='minutes'>
    /// The compass angle in minutes. From 0 to 21599
    /// </param>
    /// <param name='distanceCm'>
    /// Distance in cm.
    /// </param>
    public RelativePosition(int minutes, int distanceCm) {
        resetAngleInMinDstInCm(minutes, distanceCm);
    }

    public void resetFromTo(Vector3f newV3from, Vector3f newV3to) {
        v2fromTo = new Vector2f(newV3to.x - newV3from.x, newV3to.z - newV3from.z);
//        compassAngle = Vector2f.Angle(v2fromTo, north);
        compassAngle = v2fromTo.angleBetween(north) * 180 / FastMath.PI;
        if (compassAngle < 0) {
            compassAngle = 360 + compassAngle;
        }
//        if (v2fromTo.x > 0) {
//            compassAngle = 360 - compassAngle;
//        }
//        if (compassAngle >= 360) {
//            compassAngle = compassAngle - 360; //to avoid rounding errors (e.g. compassAngle is rounded to 0)
//        }
    }

    public void resetDgrMinDstInCm(int degrees, int minutes, int distanceCm) {
        float dst = ((float) distanceCm) / 100f;
        compassAngle = ((float) degrees) + ((float) minutes) / 60f;
        v2fromTo = new Vector2f((float) (-Math.sin(FastMath.DEG_TO_RAD * compassAngle) * dst), (float) (-Math.cos(FastMath.DEG_TO_RAD * compassAngle) * dst));
    }

    public void resetAngleInMinDstInCm(int minutes, int distanceCm) {
        float dst = ((float) distanceCm) / 100f;
        minutes = minutes % 21600; //just for sure
        compassAngle = ((float) minutes) / 60f;
        v2fromTo = new Vector2f((float) -Math.sin(FastMath.DEG_TO_RAD * compassAngle) * dst, (float) (-Math.cos(FastMath.DEG_TO_RAD * compassAngle) * dst));
    }

    public float getCompassAngle() {
        return compassAngle;
    }

    public int getCompassAngleDegrees() {
        return (int) Math.floor(compassAngle);
    }

    public int getCompassAngleMinutes() {
        return (int) Math.floor(60 * (compassAngle - Math.floor(compassAngle)));
    }

    public int getCompassAngleInMinutes() {
        return (getCompassAngleDegrees() * 60 + getCompassAngleMinutes()) % 21600;
    }

    /// <summary>
    /// Gets the distance im meters as float.
    /// </summary>
    /// <returns>
    /// The distance im meters as float.
    /// </returns>
    public float getDst() {
        return v2fromTo.length() * Measure.M;
    }

    public int getDstInCm() {
        return (int) Math.floor(v2fromTo.length() * Measure.CM);
    }

    public Vector2f getVector() {
        return v2fromTo;
    }

    /// <summary>
    /// Converts data about relative position of the cleaning robot and objects listed in livingBeings
    /// to an array of readings of cleaning bot's proximity sensor
    /// The sensor detects the proximity in 8 regions, each covering 45 deggrees.
    /// Regions are ordered clockwise.
    /// If the distance is greater as the maximum distance measurable by the sensor, it returns
    /// the maximum distance (20m=2000cm).
    /// (by SKor)
    /// </summary>
    /// 
    /// <returns>
    /// Array of the lenght 8, containing readings of cleaning bot's proximity sensor:
    /// [0] - <0,45) - NNE (north north east)
    /// [1] - <45,90) - ENE (east north east)
    /// [2] - <90,135) - ESE (east south east)
    /// [3] - <135,180) - SSE (south south east)
    /// [4] - <180, 225) - SSW (south south west)
    /// [5] - <225,270) - WSW (west south west)
    /// [6] - <270,315) - WNW 
    /// [7] - <315,360) - NNW 
    /// Readings are distances in cm
    /// </returns>
    /// 
    /// <param name='botPosition'>
    /// Cleaning robot position
    /// </param>
    /// 
    /// <param name='livingBeings'>
    /// List of game objects that represent living beings in the scene.
    /// </param>
    public static int[] readProximitySensor(List<Node> livingBeings, Vector3f botPosition) {

        //Debug.Log("CleanBotModel.readProximitySensor>> botPosition=" + botPosition);

        //First, some constants to set up the sensor
        final float maxDist = 5;	  //maximum measurable distance (v metroch)
        final float maxDist2Reg = 1; //if the distance is less than maxDist2Reg then the sensor will regard the object as occupying two neighbouring regions
        final int regSize = 45;   //size of a region in degrees



        //Second, define some auxiliary values
        final int regCount = 360 / regSize;   //number of regions
        final float rMidAngle = regSize / 2; //middle of the region (22.5 for regSize = 45)
        final float rTolerance = regSize / 12; //tolerance for the middle and the borders of the region (3.75 for regSize = 22.5)



        //Third, declare and initialize the array of distances
        float[] regionProx = new float[regCount];

        for (int i = 0; i < regionProx.length; i++) {
            regionProx[i] = maxDist;
        }


        //Fourth, fill in the array of distances
        for (Node lb : livingBeings) {
            //b2lb - relative position from bot to the living being lb
            RelativePosition b2lb = new RelativePosition(botPosition, ((AbstractCharacter) lb).getNodeForMovement().getLocalTranslation());

            int regNo = ((int) Math.floor(b2lb.getCompassAngle())) / regSize;  //region number
            float rAngle = b2lb.getCompassAngle() - regNo * regSize;  //angle within the region

            //compute bounds for when to update neighboring regions, too
            float whenPrev, whenNext;
            if (b2lb.getDst() > maxDist2Reg) { //not very close, update neighboring regions if the object is close to border
                whenPrev = rTolerance;			//rAngle in <0,rTolerance), e.g. <0,3.75)
                whenNext = regSize - rTolerance;	//rAngle in (regSize-rTolerance,regSize), e.g. (41.25,45)
            } else { //very close, always update 2 regions, except when in the middle
                whenPrev = rMidAngle - rTolerance;	//rAngle in <0,rMidAngle-rTolerance), e.g. <0,18.75)
                whenNext = rMidAngle + rTolerance;	//rAngle in (rMidAngle+rTolerance,regSize), e.g. (26.25,45)
            }

            //update this region
            if (regionProx[regNo] > b2lb.getDst()) {
                regionProx[regNo] = b2lb.getDst();
            }

            //update neighboring regions if necessary
            if (rAngle < whenPrev) { //previous region
                if (regNo == 0) {
                    if (regionProx[regCount - 1] > b2lb.getDst()) {
                        regionProx[regCount - 1] = b2lb.getDst();
                    }
                } else {
                    if (regionProx[regNo - 1] > b2lb.getDst()) {
                        regionProx[regNo - 1] = b2lb.getDst();
                    }
                }
            } else if (rAngle > whenNext) { //next region
                if (regNo == (regCount - 1)) {
                    if (regionProx[0] > b2lb.getDst()) {
                        regionProx[0] = b2lb.getDst();
                    }
                } else {
                    if (regionProx[regNo + 1] > b2lb.getDst()) {
                        regionProx[regNo + 1] = b2lb.getDst();
                    }
                }
            }

        }

        //Fifth, convert and return the array of distances in cm
        int[] regionProxInt = new int[regionProx.length];

        for (int i = 0; i < regionProxInt.length; i++) {
            regionProxInt[i] = (int) Math.floor(regionProx[i] * 100);
        }

        return regionProxInt;

    }
}